import pandas as pd
import numpy as np
from scipy import stats
from sklearn import metrics
from xgboost import XGBClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.feature_selection import chi2
from sklearn.linear_model import LinearRegression
from sklearn.feature_selection import RFE
import warnings

warnings.filterwarnings('ignore')



np.random.seed(1)

visibleSize = 18

hiddenSize = 1000

def IFS(X1_test,Y_test,X1_train,Y_train0,X_origtrain1,X_origtest1, visibleSize, classification_type):
    FeaSub1 = []
    FeaSubtrain = []
    FeaSub2 = []
    FeaSubtrain2 = []
    Y_temp = []
    Y_train = []
    Y_temp.append(list(Y_test))
    Y_train.append(list(Y_train0))
    feature_num = 0
    dic_acc = {}

    dic = {}
    acc1 = [] ;acc2 = []
    sn1 = [] ;sn2 = []
    sp1 = [] ;sp2 = []
    #tn1 = [] ;tn2 = []
    #fp1 = [] ;fp2 = []
    #fn1 = [] ;fn2 = []
    #tp1 = [] ;tp2 = []
    
    #The middle layer comes out to be 1000 features
    num = 1000
    
    #Comparison of the constructed features with the original visibleSize features
    for i in range(num):
        feature_num += 1
        FeaSub1.append(list(X1_test[:, i]))
        X_use1 = np.transpose(np.array(FeaSub1))
        FeaSubtrain.append(list(X1_train[:, i]))
        X_usetrain = np.transpose(np.array(FeaSubtrain))
        if i < visibleSize:
            FeaSub2.append(list(X_origtest1[:, i]))
            FeaSubtrain2.append(list(X_origtrain1[:, i]))
        X_use2 = np.transpose(np.array(FeaSub2))
        X_usetrain2 = np.transpose(np.array(FeaSubtrain2))
        Y = np.transpose(np.array(Y_temp))
        Y_train1 = np.transpose(np.array(Y_train))

        dic[str(feature_num)] = result(X_use1, X_usetrain, Y,Y_train1,X_use2,X_usetrain2, classification_type)

        acc1.append(dic[str(feature_num)]['acc1'])
        sn1.append(dic[str(feature_num)]['sn1'])
        sp1.append(dic[str(feature_num)]['sp1'])
        acc2.append(dic[str(feature_num)]['acc2'])
        sn2.append(dic[str(feature_num)]['sn2'])
        sp2.append(dic[str(feature_num)]['sp2'])
        #tn.append(dic[str(feature_num)]['tn'])
        #fp.append(dic[str(feature_num)]['fp'])
        #fn.append(dic[str(feature_num)]['fn'])
        #tp.append(dic[str(feature_num)]['tp'])

        dic_acc[str(feature_num)] = {'acc1': dic[str(feature_num)]['acc1'],
                                     'sn1': dic[str(feature_num)]['sn1'],
                                     'sp1': dic[str(feature_num)]['sp1'],
                                     'acc2': dic[str(feature_num)]['acc2'],
                                     'sn2': dic[str(feature_num)]['sn2'],
                                     'sp2': dic[str(feature_num)]['sp2']
                                     }

    dic_acc['max'] = {'acc1': max(acc1),
                      'sn1': sn1[acc1.index(max(acc1))],
                      'sp1': sp1[acc1.index(max(acc1))],
                      'acc2': max(acc2),
                      'sn2': sn1[acc2.index(max(acc2))],
                      'sp2': sp1[acc2.index(max(acc2))]
                      }

    return dic_acc


def result(X_use1, X_usetrain, Y,Y_train1,X_use2,X_usetrain2, classification_type):
    model = get_clf(classification_type)

    model.fit(X_usetrain,Y_train1.ravel())
    y_pred1 = model.predict(X_use1)
    tn1, fp1, fn1, tp1 = metrics.confusion_matrix(Y, y_pred1,labels=[0,1]).ravel()

    model.fit(X_usetrain2,Y_train1.ravel())
    y_pred2 = model.predict(X_use2)
    tn2, fp2, fn2, tp2 = metrics.confusion_matrix(Y, y_pred2,labels=[0,1]).ravel()
    
    acc1 = (tp1 + tn1) / float(tp1 + fn1 + tn1 + fp1)
    sn1 = tp1 / float(tp1 + fn1) 
    sp1 = tn1 / float(tn1 + fp1)
    acc2 = (tp2 + tn2) / float(tp2 + fn2 + tn2 + fp2)
    sn2 = tp2 / float(tp2 + fn2) 
    sp2 = tn2 / float(tn2 + fp2)
    result = {'acc1': acc1, 'sn1': sn1, 'sp1': sp1, 'acc2': acc2, 'sn2': sn2, 'sp2': sp2}
    return result

def Ttest(X, Y, ttest_num):
    Y1 = np.where(Y == 1)[0]
    Y0 = np.where(Y == 0)[0]
    p_list = []
    for i in range(len(X[0])):
        p_l = stats.levene(X[Y1, i], X[Y0, i])[1]
        equal_var = [True if p_l > 0.05 else False]
        p_list.append(stats.ttest_ind(X[Y1, i], X[Y0, i], equal_var=equal_var)[1])
    T_index = np.argsort(np.array(p_list))[0:ttest_num]
    X1 = X[:, T_index]
    return T_index


def Wtest(X, Y, wtest_num):
    Y1 = np.where(Y == 1)[0]
    Y0 = np.where(Y == 0)[0]
    p_list = []
    for i in range(len(X[0])):
        p_list.append(stats.ranksums(X[Y1, i], X[Y0, i])[1])
    W_index = np.argsort(np.array(p_list))[0:wtest_num]
    return W_index


def Chi2(X, Y, ctest_num):
    p_list = chi2(X, Y)[1]
    C_index = np.argsort(np.array(p_list))[0:ctest_num]
    return C_index


def RF(X, Y, rtest_num):
    forest = RandomForestClassifier(random_state=0, n_jobs=1)
    forest.fit(X, Y)
    importance = forest.feature_importances_
    R_index = np.argsort(np.array(importance))[::-1][0:rtest_num]
    return R_index


def LR_RFE(X, Y, lrtest_num):
    clf = LinearRegression()
    rfe = RFE(clf, n_features_to_select=1)
    rfe.fit(X, Y)
    rank = rfe.ranking_
    LR_index = np.argsort(np.array(rank))[0:lrtest_num]
    return LR_index


def SVM_RFE(X, Y, srtest_num):
    clf = SVC(kernel='linear', random_state=0)
    rfe = RFE(clf, n_features_to_select=1)
    rfe.fit(X, Y)
    rank = rfe.ranking_
    SR_index = np.argsort(np.array(rank))[0:srtest_num]
    return SR_index

def get_clf(clf_name):
    if clf_name == 'DTree':
        model = DecisionTreeClassifier(random_state=0)
    elif clf_name == 'KNN':
        model = KNeighborsClassifier()
    elif clf_name == 'LogisticR':
        model = LogisticRegression(random_state=0)
    elif clf_name == 'NBayes':
        model = GaussianNB()
    elif clf_name == 'RF':
        model = RandomForestClassifier(random_state=0)
    elif clf_name == 'SVM':
        model = SVC(kernel='linear', random_state=0)
    elif clf_name == 'XGB':
        model = XGBClassifier(random_state=0)
    else:
        pass
    return model


def predict(feature_selection_type, classification_type, test_data, test_label ,train_data,train_label):
    
    train_sae = pd.read_csv("./result/" + feature_selection_type + "_train_sae.csv", index_col=0)
    test_sae = pd.read_csv("./result/" + feature_selection_type + "_test_sae.csv", index_col=0)
    
    X_train = np.array(train_sae)
    X_test = np.array(test_sae)
    X_origtest = np.array(test_data)
    X_origtrain = np.array(train_data)

    D = []
    for i in test_label['0'].tolist():
        D.append(1 if i == 1 else 0)
    Y_test = np.array(D)
    
    C = []
    for i in train_label['0'].tolist():
        C.append(1 if i == 1 else 0)
    Y_train = np.array(C)
    print(Y_train)
    #Quiz for label size

    index = pd.read_csv("./result/" + feature_selection_type + "_index.csv", index_col=0)["0"].tolist()

    X1_test = X_test[:, index]
    X1_train = X_train[:, index]
    
    
    
    if feature_selection_type == 'Ttest':
        index1 = Ttest(X_origtrain, Y_train, visibleSize).tolist()
    elif feature_selection_type == 'Wtest':
        index1 = Wtest(X_origtrain, Y_train, visibleSize).tolist()
    elif feature_selection_type == 'Chi2':
        index1 = Chi2(X_origtrain, Y_train, visibleSize).tolist()
    elif feature_selection_type == 'RF':
        index1 = RF(X_origtrain, Y_train, visibleSize).tolist()
    elif feature_selection_type == 'LR_RFE':
        index1 = LR_RFE(X_origtrain, Y_train, visibleSize).tolist()
    elif feature_selection_type == 'SVM_RFE':
        index1 = SVM_RFE(X_origtrain, Y_train, visibleSize).tolist()
    
    X_origtrain1 = X_origtrain[:,index1]
    X_origtest1 = X_origtest[:,index1]
    dic_acc = IFS(X1_test,Y_test,X1_train,Y_train,X_origtrain1,X_origtest1, visibleSize, classification_type)
    df_acc = pd.DataFrame(dic_acc)
    df_acc.to_csv("./result/" + feature_selection_type + "_" + classification_type + "_result.csv")


if __name__ == '__main__':
    feature_selection_type = "Wtest"
    classification_type = "NBayes"
    train_data = pd.read_csv("./result/sample/data_train.csv", index_col=0)
    train_label = pd.read_csv("./result/sample/label_train.csv", index_col=0)
    test_data = pd.read_csv("./result/sample/data_test.csv", index_col=0)
    test_label = pd.read_csv("./result/sample/label_test.csv", index_col=0)
    predict(feature_selection_type, classification_type, test_data, test_label ,train_data,train_label)
